import os
import ast
import sys
import subprocess
from traceback import print_exc
sys.path.insert(0, os.path.abspath(os.pardir))
from pydetector.astexport import export_dict # noqa: E402

__all__ = ['check_ast']

# This will store the AST to use for the "other" Python. It will be
# loaded the first time its tested by getOtherPythonAstEngine
OTHERPYTHON_AST = None

if sys.version_info[0] not in (2, 3):
    raise Exception('Sorry, only python 2 and 3 is supported')
PYMAJOR_CURRENT = sys.version_info[0]
PYMAJOR_OTHER = 2 if PYMAJOR_CURRENT == 3 else 3

def check_ast(code, try_other_on_sucess=False, verbosity=0,
              py2_exec='python2', py3_exec='python3'):
    """
    Try with the ast.parse of both Python 2 and 3 and then
    iterate over the retrieved AST to find specific syntax elements.

    Args:
        code (str) the source code to extract the AST for.

        try_other_on_sucess (bool): if the AST parser of the Python version that is
        running this script work try anyway with the other interpreter (True) or
        not (False). Please note that enabling this will usually increase a lot
        the "detection" rate of files as the version of the current interpreter.

        verbosity (int): level from 0 to 2. > 1 will show exceptions when parsing
        the AST.

        py2_exec (str): path or name (if in PATH) of the Python 2 interpreter when running
        this under Python 3.

        py3_exec (str): path or name (if in PATH) of the Python 3 interpreter to use when
        running this under Python 2.
    """

    current_ok = other_ok = False
    current_ast = other_ast = None
    pyexec_other = py2_exec if PYMAJOR_OTHER == 2 else py3_exec

    try:
        current_ast = export_dict(code)
        current_ok = True
    except:
        # current_ok remains false
        if verbosity > 1:
            print('>>>> ASTCHECK: exception while parsing AST with Python%d:'
                    % PYMAJOR_CURRENT)
            print_exc()
            print('<<<< exception output end')

    if verbosity:
        print('AST extractable with version %d?: %s' % (PYMAJOR_CURRENT, str(current_ok)))

    if not current_ok or try_other_on_sucess:
        # Write a tempfile with the code for the external interpreter
        cmd = [pyexec_other, "-c",
                "import ast,pydetector.astexport,sys;"
                "r=sys.stdin.read();"
                "print(pydetector.astexport.export_dict(r))"]

        if verbosity > 1:
            print('Running in other Python:\n%s' % ' '.join(cmd))

        err = None
        try:
            p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate(code.encode('utf-8'))
            if p.returncode == 0:
                if PYMAJOR_CURRENT == 3:
                    out = out.decode('utf-8')
                other_ast = ast.literal_eval(out)
                other_ok = True
            elif verbosity > 1:
                print('>>>> ASTCHECK: error while parsing AST with Python%d:'
                        % PYMAJOR_OTHER)
                print(err)
                print('<<<< error output end')
        except:
            other_ok = False
            if verbosity > 1:
                print('>>>> ASTCHECK: exception while parsing AST with Python%d:'
                        % PYMAJOR_OTHER)
                print_exc()
                print('<<<< exception output end')

    if verbosity:
        print('AST extractable with version %d?: %s' % (PYMAJOR_OTHER, str(other_ok)))

    if PYMAJOR_CURRENT == 2:
        py2_ast = current_ast
        py3_ast = other_ast
    else:
        py3_ast = current_ast
        py2_ast = other_ast

    if current_ok and not other_ok:
        version = PYMAJOR_CURRENT
    elif other_ok and not current_ok:
        version = PYMAJOR_OTHER
    elif current_ok and other_ok:
        version = 6
    elif not current_ok and not other_ok:
        version = 0

    return version, py2_ast, py3_ast



import re
import six

__all__ = ['check_syntax_regex', 'check_modules_regex', 'check_modulesymbols_regex']

# Syntactic elements. Second item in the tuple is the score.
# TODO: test if I can simplify the regexes while passing tests to improve speed

LINESTART          = r"(^|;)\s*"
WHITEORSEP         = r"(^|\s|;|:|,|=)+"
PARENTH_ARGS       = r"\s*\(.*\)"
WHITEORSEPORPARENS = r"(^|\s|;|:|,|=|%s)+" % PARENTH_ARGS

PY3OTHER_REGEXP = [
    (re.compile(WHITEORSEP + r"raise\s+.*\s+from\s+None", re.MULTILINE), 100),
    (re.compile(WHITEORSEP + r"nonlocal\s+", re.MULTILINE), 100),
]

PY2OTHER_REGEXP = [
    (re.compile(WHITEORSEP + r"unicode"    + PARENTH_ARGS, re.MULTILINE), 25),
    (re.compile(r"\w\.iterkeys"            + PARENTH_ARGS, re.MULTILINE), 25),
    (re.compile(r"\w\.iteritems"           + PARENTH_ARGS, re.MULTILINE), 25),
    (re.compile(r"\w\.itervalues"          + PARENTH_ARGS, re.MULTILINE), 25),
    (re.compile(r"\w\.viewkeys"            + PARENTH_ARGS, re.MULTILINE), 25),
    (re.compile(r"\w\.viewitems"           + PARENTH_ARGS, re.MULTILINE), 25),
    (re.compile(r"\w\.viewvalues"          + PARENTH_ARGS, re.MULTILINE), 25),
    (re.compile(WHITEORSEP + r"xrange"     + PARENTH_ARGS, re.MULTILINE), 100),
    (re.compile(WHITEORSEP + r"xreadlines" + PARENTH_ARGS, re.MULTILINE), 100),
    (re.compile(WHITEORSEP + r"raw_input"  + PARENTH_ARGS, re.MULTILINE), 100),
    (re.compile(WHITEORSEP + r"basestring\s*:*[^\(]",      re.MULTILINE), 100),
    (re.compile(WHITEORSEP + r"print\s+[^\(]",             re.MULTILINE), 100),
    (re.compile(WHITEORSEP + r"__metaclass__\s*=",         re.MULTILINE), 100),
    (re.compile(WHITEORSEP + r"raise\s+\w+(\.\w+)?,\s*",   re.MULTILINE), 100),
    (re.compile(r"\w\.has_key"             + PARENTH_ARGS, re.MULTILINE), 100),
]


PY2ONLY_MODULES_REGEXP = None
PY3ONLY_MODULES_REGEXP = None
def generate_modules_regex():
    global PY2ONLY_MODULES_REGEXP
    global PY3ONLY_MODULES_REGEXP

    py2only_modules = [
        "cfmfile", r"[^hashlib\.]([^_]md5|[^_]sha)", "mimetools", "MimeWriter", "mimify",
        "(multi|posix)file", "rfc822", "timing", "audiodev", "stringold", "bsddb185", "Canvas",
        "commands", "compiler", "dircache", "fpformat", "(html|mh|sgml|cookie|xmlrpc|http)lib",
        "imageop", "linuxaudiodev", "c?StringIO", "cPickle",
        "popen2", "sre", "statvfs", r"[^collections\.]User(Dict|String|List)",
        "(Config|HTML|robot)(P|p)arser", "copy_reg", "Queue", "SockerServer", "sets",
        "dbhash", "(g|dumb|which|any)dbm", "(Doc|Simple)XMLRPCServer",
        "Cookie", "(Base|Simple|CGI)HTTPServer", "urlparse"

    ]
    py3only_modules = [
        "configparser", "copyreg", "queue", "socketserver", "ipaddress", "lzma",
        "(context|repr)lib", "six", # six usually mean both compatible, so just mark as 3
        "dbm\.(bsd|dumb|ndbm|gdbm)", "xmlrpc\.(client|server)", "http\.(client|server)",
        "urllib\.(parse|robotparser)"
    ]

    def import_regex_gen(modlist):
        strregex_import = LINESTART + r"import\s+.*(%s)(\s*|,|$)" % "|".join(modlist)
        strregex_from   = LINESTART + r"from\s+(%s)\s+import\s+" % "|".join(modlist)
        return re.compile(strregex_import + "|" + strregex_from, re.MULTILINE)

    PY2ONLY_MODULES_REGEXP = import_regex_gen(py2only_modules)
    PY3ONLY_MODULES_REGEXP = import_regex_gen(py3only_modules)


PY3MODULESYMBOLS_REGEXPS = []
PY2MODULESYMBOLS_REGEXPS = []
def generate_modulesymbols_regex():
    global PY3MODULESYMBOLS_REGEXPS
    global PY2MODULESYMBOLS_REGEXPS

    py2only_modulesymbols = {
        "os":       ["getcwdu"],
        "sys":      ["exitfunc", "maxint", "exc_type", "exc_value", "exc_traceback"],
        "operator": ["isCallable", "sequenceIncludes", "isSequenceType",
                    "isMappingType", "isNumberType", "repeat", "irepeat"],
    }

    py3only_modulesymbols = {
        "abc":             ["get_cache_token"],
        "contextlib":      ["supress"],
        "filecmp":         ["clear_cache"],
        "functools":       ["partialmethod"],
        "gc":              ["get_stats"],
        "glob":            ["scape"],
        "hashlib":         ["pbkdf2_hmac"],
        "html":            ["unscape"],
        "inspect":         ["signature", "Parameter", "BoundArguments", "getclosurevars"],
        "multiprocessing": ["spawn", "forkserver"],
        "operator":        ["length_hint"],
        "os":              ["get_inheritable", "set_inheritable", "get_handle_inheritable",
                            "set_handle_inheritable", "sendfile", "pipe2", "getcwdu"],
        r"past\.utils":    ["old_div"],
        "poplib":          ["capa", "stsl"],
        "re":              ["fullmatch"],
        "resource":        ["prlimit"],
        "shutil":          ["disk_usage"],
        "signal":          ["pthread_sigmask", "pthread_kill", "sigpending", "sigwait",
                           "sigwaitinfo", "sigtimedwait"],
        "socket":          ["sendmsg", "recvmsg", "recvmsg_info"],
        "ssl":             ["create_default_context", "get_default_verify_paths"],
        "struct":          ["iter_unpack"],
        "sys":             ["getallocatedblocks", "implementation", "maxsize", "exc_info"],
        "textwrap":        ["indent"],
        "time":            ["get_clock_info", "monotonic", "perf_counter", "process_time"],
        "traceback":       ["clear_frames"],
        "types":           ["MappingProxyType", "new_class", "prepare_class"],
        "weakref":         ["WeakMethod", "finalize"],
        r"xml\.etree":     ["XMLPullParser"],
        "zlib":            ["ZLIB_RUNTIME_VERSION"],
    }

    strregex_import = r"^\s*from\s+%s\s+import\s+%s(\s+|,|$).*"
    strregex_usage  = WHITEORSEP + r"%s\.%s" + WHITEORSEPORPARENS

    for modname, symbollist in six.iteritems(py3only_modulesymbols):
        PY3MODULESYMBOLS_REGEXPS.append(re.compile(
            strregex_import % (modname, "|".join(symbollist)) + "|" +
            strregex_usage  % (modname, "|".join(symbollist)),
            re.MULTILINE
        ))

    for modname, symbollist in six.iteritems(py2only_modulesymbols):
        PY2MODULESYMBOLS_REGEXPS.append(re.compile(
            strregex_import % (modname, "|".join(symbollist)) + "|" +
            strregex_usage  % (modname, "|".join(symbollist)),
            re.MULTILINE
        ))


# Generate compiled regexes at import time
generate_modules_regex()
generate_modulesymbols_regex()


def check_syntax_regex(code, matches):
    """
    Test for syntax elements specific of some Python version.

    Args:
        code (str): The code
        matches (List[Tuple[str, str]]): the list of matching rules. It will
        be modified in-place

    Returns:
        A tuple with the py3_score and the py2_score
    """
    py2_score = py3_score = 0

    for regtuple in PY3OTHER_REGEXP:
        m = regtuple[0].findall(code)
        if m:
            py3_score += (regtuple[1] * len(m))
            matches.append(("PY3SYNTAX_"+regtuple[0].pattern, m))

    for regtuple in PY2OTHER_REGEXP:
        m = regtuple[0].findall(code)
        if m:
            py2_score += (regtuple[1] * len(m))
            matches.append(("PY2SYNTAX_"+regtuple[0].pattern, m))

    return py2_score, py3_score

def check_modules_regex(code, matches, match_score=100):
    """
    Test for modules specific of some Python version.

    Args:
        code (str): The code
        matches (List[Tuple[str, str]]): the list of matching rules. It will
        be modified in-place

    Returns:
        A tuple with the py3_score and the py2_score
    """
    py2_score = py3_score = 0

    m = PY3ONLY_MODULES_REGEXP.findall(code)
    if m:
        py3_score += (match_score * len(m))
        for match in m:
            matches.append(('PY3MODS', match))

    m = PY2ONLY_MODULES_REGEXP.findall(code)
    if m:
        py2_score += (match_score * len(m))
        for match in m:
            matches.append(('PY2MODS', match))

    return py2_score, py3_score


def check_modulesymbols_regex(code, matches, symbols_score=100):
    """
    Test for module symbols specific of some Python version. Please note
    that this test can be very slow compared with the others since a
    lot of regular expressions are tested.

    Args:
        code (str): The code
        matches (List[Tuple[str, str]]): the list of matching rules. It will
        be modified in-place

    Returns:
        A tuple with the py3_score and the py2_score
    """
    py2_score = py3_score = 0
    for symregex in PY3MODULESYMBOLS_REGEXPS:
        m = symregex.findall(code)
        if m:
            matches.append(('PY3SYMS:' + symregex.pattern, m))
            py3_score += (symbols_score * len(m))

    # Currently this doesn't test for any py2symbols
    return py2_score, py3_score



# This module uses (modified) code from:
# https://github.com/fpoli/python-astexport/blob/master/astexport/export.py
# Hosted at: https://github.com/edreamleo/python-to-coffeescript

# Released under the MIT License.

import ast
import sys
import json
import types
import tokenize
import token as token_module

isPython3 = sys.version_info >= (3, 0, 0)

def export_dict(codestr):
    visitor = DictExportVisitor(codestr)
    return visitor.parse()

def export_json(codestr, pretty_print=False):
    return json.dumps(
        export_dict(codestr),
        indent=2 if pretty_print else 0,
        ensure_ascii=False
    )

try:
    import msgpack
except ImportError:
    pass
else:
    def export_msgpack(codestr):
        return msgpack.packb(export_dict(codestr))


class TokenSync(object):
    '''A class to sync and remember tokens.'''

    def __init__(self, codestr):
        '''Ctor for TokenSync class.'''
        tokens = list(
                tokenize.generate_tokens(ReadLinesClass(codestr).next)
        )

        self.s = codestr
        self.first_leading_line = None
        self.lines = [z.rstrip() for z in TokenSync.splitLines(codestr)]
        # Order is important from here on...
        self.nl_token = self.make_nl_token()
        self.line_tokens = self.make_line_tokens(tokens)
        self.blank_lines = self.make_blank_lines()
        self.string_tokens = self.make_string_tokens()
        self.ignored_lines = self.make_ignored_lines()

    @staticmethod
    def splitLines(s):
        '''Split s into lines, preserving trailing newlines.'''
        return s.splitlines(True) if s else []

    @staticmethod
    def isString(s):
        '''Return True if s is any string, but not bytes.'''
        if isPython3:
            return isinstance(s, str)
        else:
            return isinstance(s, types.StringTypes)

    @staticmethod
    def isUnicode(s):
        '''Return True if s is a unicode string.'''
        # pylint: disable=no-member
        if isPython3:
            return isinstance(s, str)
        else:
            return isinstance(s, types.UnicodeType)

    @staticmethod
    def toUnicode(s, encoding='utf-8', reportErrors=False):
        '''Connvert a non-unicode string with the given encoding to unicode.'''
        if isPython3:
            def u(s):
                return s

            def ue(s, encoding):
                return s if TokenSync.isUnicode(s) else str(s, encoding)
        else:
            def u(s):
                return unicode(s) # noqa: F821

            def ue(s, encoding):
                return unicode(s, encoding) # noqa: F821

        trace = False
        if TokenSync.isUnicode(s):
            return s
        if not encoding:
            encoding = 'utf-8'
        try:
            s = s.decode(encoding, 'strict')
        except UnicodeError:
            s = s.decode(encoding, 'replace')
            if trace or reportErrors:
                print("toUnicode: Error converting %s... from %s encoding to unicode" % (
                    s[: 200], encoding))
        except AttributeError:
            if trace:
                print('toUnicode: AttributeError!: %s' % s)
            s = u(s)
        if trace and encoding == 'cp1252':
            print('toUnicode: returns %s' % s)
        return s

    def make_blank_lines(self):
        '''Return of list of line numbers of blank lines.'''
        result = []
        for i, aList in enumerate(self.line_tokens):
            # if any([self.token_kind(z) == 'nl' for z in aList]):
            if len(aList) == 1 and self.token_kind(aList[0]) == 'nl':
                result.append(i)
        return result

    def make_ignored_lines(self):
        '''
        Return a copy of line_tokens containing ignored lines,
        that is, full-line comments or blank lines.
        These are the lines returned by leading_lines().
        '''
        result = []
        for i, aList in enumerate(self.line_tokens):
            for z in aList:
                if self.is_line_comment(z):
                    result.append(z)
                    break
            else:
                if i in self.blank_lines:
                    result.append(self.nl_token)
                else:
                    result.append(None)
        assert len(result) == len(self.line_tokens)
        for i, aList in enumerate(result):
            if aList:
                self.first_leading_line = i
                break
        else:
            self.first_leading_line = len(result)
        return result

    def make_line_tokens(self, tokens):
        '''
        Return a list of lists of tokens for each list in self.lines.
        The strings in self.lines may end in a backslash, so care is needed.
        '''
        n, result = len(self.lines), []
        for i in range(0, n+1):
            result.append([])
        for token in tokens:
            t1, t2, t3, t4, t5 = token
            kind = token_module.tok_name[t1].lower()
            srow, scol = t3
            erow, ecol = t4
            line = erow-1 if kind == 'string' else srow-1
            result[line].append(token)
        assert len(self.lines) + 1 == len(result), len(result)
        return result

    def make_nl_token(self):
        '''Return a newline token with '\n' as both val and raw_val.'''
        t1 = token_module.NEWLINE
        t2 = '\n'
        t3 = (0, 0) # Not used.
        t4 = (0, 0) # Not used.
        t5 = '\n'
        return t1, t2, t3, t4, t5

    def make_string_tokens(self):
        '''Return a copy of line_tokens containing only string tokens.'''
        result = []
        for aList in self.line_tokens:
            result.append([z for z in aList if self.token_kind(z) == 'string'])
        assert len(result) == len(self.line_tokens)
        return result

    # TODO: call this check?
    def check_strings(self):
        '''Check that all strings have been consumed.'''
        for i, aList in enumerate(self.string_tokens):
            if aList:
                for z in aList:
                    print(self.dump_token(z))

    def dump_token(self, token, verbose=False):
        '''Dump the token. It is either a string or a 5-tuple.'''
        if TokenSync.isString(token):
            return token
        else:
            t1, t2, t3, t4, t5 = token
            kind = TokenSync.toUnicode(token_module.tok_name[t1].lower())
            val = TokenSync.toUnicode(t2)
            if verbose:
                return 'token: %10s %r' % (kind, val)
            else:
                return val

    def is_line_comment(self, token):
        '''Return True if the token represents a full-line comment.'''
        t1, t2, t3, t4, t5 = token
        kind = token_module.tok_name[t1].lower()
        raw_val = t5
        return kind == 'comment' and raw_val.lstrip().startswith('#')

    def leading_lines(self, node):
        '''Return a list of the preceding comment and blank lines'''
        # This can be called on arbitrary nodes.
        leading = []
        if hasattr(node, 'lineno'):
            i, n = self.first_leading_line, node.lineno
            while i < n:
                token = self.ignored_lines[i]
                if token:
                    s = self.token_raw_val(token).rstrip()+'\n'
                    leading.append(s)
                i += 1
            self.first_leading_line = i
        return leading

    def leading_string(self, node):
        '''Return a string containing all lines preceding node.'''
        return ''.join(self.leading_lines(node))

    def token_kind(self, token):
        '''Return the token's type.'''
        t1, t2, t3, t4, t5 = token
        return TokenSync.toUnicode(token_module.tok_name[t1].lower())

    def token_raw_val(self, token):
        '''Return the value of the token.'''
        t1, t2, t3, t4, t5 = token
        return TokenSync.toUnicode(t5)

    def token_val(self, token):
        '''Return the raw value of the token.'''
        t1, t2, t3, t4, t5 = token
        return TokenSync.toUnicode(t2)

    def trailing_comment(self, node):
        '''
        Return a string containing the trailing comment for the node, if any.
        The string always ends with a newline.
        '''
        if hasattr(node, 'lineno'):
            return self.trailing_comment_at_lineno(node.lineno)
        else:
            return '\n'

    def trailing_comment_at_lineno(self, lineno):
        '''Return any trailing comment at the given node.lineno.'''
        tokens = self.line_tokens[lineno-1]
        for token in tokens:
            if self.token_kind(token) == 'comment':
                raw_val = self.token_raw_val(token).rstrip()
                if not raw_val.strip().startswith('#'):
                    val = self.token_val(token).rstrip()
                    s = ' %s\n' % val
                    return s
        return '\n'

    def trailing_lines(self):
        '''return any remaining ignored lines.'''
        trailing = []
        i = self.first_leading_line
        while i < len(self.ignored_lines):
            token = self.ignored_lines[i]
            if token:
                s = self.token_raw_val(token).rstrip()+'\n'
                trailing.append(s)
            i += 1
        self.first_leading_line = i
        return trailing

class ReadLinesClass:
    """A class whose next method provides a readline method for Python's tokenize module."""

    def __init__(self, s):
        self.lines = s.splitlines(True) if s else []
        self.i = 0

    def next(self):
        if self.i < len(self.lines):
            line = self.lines[self.i]
            self.i += 1
        else:
            line = ''
        return line

    __next__ = next

class DictExportVisitor(object):
    ast_type_field = "ast_type"

    def __init__(self, codestr, ast_parser = ast.parse, tsync_class=TokenSync):
        """
        Initialize the Token Syncer composited object, parse the source code
        and start visiting the node tree to add comments and other modifications

        Args:
            codestr (string): the string with the source code to parse and visit.

            ast_parser (function, optional): the AST parser function to use. It needs to take
            a string with the code as parameter. By default it will be ast.parse from stdlib.

            tsync_class (class, optional): the class to use to sinchronize the tokenizer with
            the AST visits. This is needed to extract aditional info like comments that most
            AST parsers doesn't include in the tree. This class need to provide the public
            methods "leading_lines", "leading_string", "trailing_comment", "trailing_lines"
            and "sync_strings". By default astexport.TokenSync will be used.
        """
        # Instantiate the TSyncClass
        self.codestr = codestr
        self.sync    = tsync_class(codestr)

    def parse(self):
        node = ast.parse(self.codestr, mode='exec')
        res = self.visit(node)
        # self.sync.check_strings()
        return res

    def visit(self, node):
        node_type = node.__class__.__name__
        # print('XXX node_type: {}'.format(node_type))
        meth = getattr(self, "visit_" + node_type, self.default_visit)
        return meth(node)

    def default_visit(self, node):
        node_type = node.__class__.__name__
        # print('XXX node_type: {}'.format(node_type))
        # Add node type
        args = {
            self.ast_type_field: node_type
        }

        # Visit fields
        for field in node._fields:
            meth = getattr(
                self, "visit_" + node_type,
                self.default_visit_field
            )
            args[field] = meth(getattr(node, field))

        # Visit attributes
        for attr in node._attributes:
            meth = getattr(
                self, "visit_" + node_type + "_" + attr,
                self.default_visit_field
            )
            # Use None as default when lineno/col_offset are not set
            args[attr] = meth(getattr(node, attr, None))
        return args

    def default_visit_field(self, node):
        node_type = node.__class__.__name__
        # print('XXX [visit_field] node_type: {}'.format(node_type))

        if isinstance(node, ast.AST):
            return self.visit(node)
        elif isinstance(node, list) or isinstance(node, tuple):
            return [self.visit(x) for x in node]
        else:
            return node

    def visit_str(self, node):
        node_type = node.__class__.__name__
        # print('XXX [visit_str] node_type: {}'.format(node_type))
        return self.visit_Str(node)

    def visit_Str(self, node):
        node_type = node.__class__.__name__
        # print('XXX [visit_Str] node_type: {}'.format(node_type))
        return {
                self.ast_type_field: "str",
                "s": node.s
        }

    def visit_Bytes(self, node):
        node_type = node.__class__.__name__
        # print('XXX [visit_Bytes] node_type: {}'.format(node_type))
        return {
                self.ast_type_field: "bytes",
                "s": node.s.decode()
        }

    def visit_NoneType(self, node):
        node_type = node.__class__.__name__
        # print('XXX [visit_NoneType] node_type: {}'.format(node_type))
        return 'None'

    # TODO: fix this
    # def visit_Name(self, node):
        # return self.visit_NameConstant(node)

    def visit_NameConstant(self, node):
        node_type = node.__class__.__name__
        # print('XXX [visit_field_NameConstant] node_type: {}'.format(node_type))
        if hasattr(node, 'value') and repr(node.value) in ('True', 'False'):
            return {
                    self.ast_type_field: "bool",
                    "b": node.value
            }
        return str(node)

    def visit_Num(self, node):
        node_type = node.__class__.__name__
        # print('XXX [visit_field_Num] node_type: {}'.format(node_type))

        if isinstance(node, int):
            return {
                self.ast_type_field: "int",
                "n": node
            }
        elif isinstance(node, float):
            return {
                self.ast_type_field: "float",
                "n": node
            }
        elif isinstance(node, complex):
            return {
                self.ast_type_field: "complex",
                "n": node.real,
                "i": node.imag
            }


if __name__ == '__main__':
    import sys
    from pprint import pprint
    f = sys.argv[1]

    with open(f) as codefile:
        pprint(export_dict(codefile.read()))
